"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const network_actions_1 = require("./network/network-actions");
const ws_1 = __importDefault(require("ws"));
const events_1 = require("events");
const collection_1 = __importDefault(require("./collections/collection"));
const client_fetch_1 = __importDefault(require("./client-fetch"));
const interceptor_manager_1 = __importDefault(require("./interceptor-manager"));
const master_1 = __importDefault(require("./interceptors/master"));
/**
 * @extends EventEmitter
 */
class Client extends events_1.EventEmitter {
    /**
     * @param {string} url The url of the server
     */
    constructor(url) {
        super();
        this.fetch = new client_fetch_1.default(this);
        // TODO: This should be on context, emit events instead of hard-coded here?
        this.users = new collection_1.default();
        this.members = new collection_1.default();
        this.messages = new collection_1.default();
        this.guilds = new collection_1.default();
        this.url = url;
        this.loggedIn = false;
        this.registerInterceptors();
    }
    registerInterceptors() {
        // TODO: How are interceptors also getting registered on server?
        master_1.default;
    }
    /**
     * @return {boolean} Whether the socket is open and connected
     */
    get connected() {
        return this.socket !== undefined && this.socket.readyState === this.socket.OPEN;
    }
    getGuilds() {
        return this.guilds;
    }
    setGuilds(guilds) {
        this.guilds = guilds;
        return this;
    }
    setGuild(id, guild) {
        this.guilds.set(id, guild);
        return this;
    }
    getUser(id) {
        return this.users.get(id) || null;
    }
    getUsers() {
        return this.users;
    }
    getAllMembers() {
        return this.members;
    }
    getMembers(guild) {
        return this.members.get(guild) || null;
    }
    getMember(guild, id) {
        // TODO
        return null;
    }
    getMessages() {
        return this.messages;
    }
    setMembers(guild, members) {
        this.members.set(guild, members);
        return this;
    }
    /**
     * @return {boolean} Whether the client is logged in
     */
    isLoggedIn() {
        return this.loggedIn;
    }
    /**
     * @return {User | undefined}
     */
    getMe() {
        return this.me;
    }
    setMe(me) {
        this.me = me;
        return this;
    }
    setToken(token, loggedIn) {
        this.token = token;
        if (loggedIn) {
            this.loggedIn = true;
        }
    }
    addMessage(message) {
        this.messages.set(message.id, message);
        return this;
    }
    useInvite(code) {
        this.send({
            _type: network_actions_1.NetworkPacketType.UseInvite,
            _auth: this.getAuth(),
            code: code
        });
        return this;
    }
    /**
     * Create the socket and setup events
     * @return {this}
     */
    connect() {
        this.socket = new ws_1.default(`ws://${this.url}`);
        this.socket.on("open", () => {
            this.emit("connected");
        });
        this.socket.on("message", async (data) => {
            this.emit("serverData", data);
            const packet = JSON.parse(data);
            if (!Client.validate(packet)) {
                this.emit("invalidPacket", packet);
                return;
            }
            // Intercept certain packets for internal use
            const interceptor = interceptor_manager_1.default.getInterceptor(packet._type);
            if (interceptor) {
                let result = interceptor(packet, this);
                if (result instanceof Promise) {
                    result = await result;
                }
                this.emit(network_actions_1.getPacketEventName(packet._type), result);
                return;
            }
            this.emit(network_actions_1.getPacketEventName(packet._type), packet);
        });
        this.socket.on("close", () => {
            this.loggedIn = false;
            this.socket = undefined;
            this.me = undefined;
            this.users = new collection_1.default();
            this.messages = new collection_1.default();
            this.guilds = new collection_1.default();
            this.emit("disconnect");
        });
        this.socket.on("error", (error) => {
            this.emit("error", error);
        });
        return this;
    }
    /**
     * Send data to the currently opened socket
     * @param {*} data The data to send
     * @return {this}
     */
    send(data) {
        if (!this.socket) {
            throw new Error("[Client.login] Not connected to remote server");
        }
        this.socket.send(JSON.stringify(data));
        return this;
    }
    /**
     * Login with credentials and request token
     * @param {UserCredentials} credentials
     * @return {this}
     */
    login(credentials) {
        // Set username for future use, gets overwrite if login fails (next login)
        this.username = credentials.username;
        this.send({
            _type: network_actions_1.NetworkPacketType.GetToken,
            credentials: credentials
        });
        return this;
    }
    /**
     * @param {string} text
     * @return {this}
     */
    message(text) {
        this.send({
            _type: network_actions_1.NetworkPacketType.MessageSend,
            _auth: this.getAuth(),
            text: text,
            channel: "443434" // TODO
        });
        return this;
    }
    /**
     * @return {ActionAuth}
     */
    getAuth() {
        if (!this.token || !this.username) {
            throw new Error("[Client.getAuth] Unexpected token or username to be null or undefined");
        }
        return {
            token: this.token,
            username: this.username
        };
    }
    /**
     *
     * @param {*} data
     * @return {boolean} Whether the data is valid
     */
    static validate(data) {
        return data !== undefined && data._type !== undefined;
    }
}
exports.default = Client;
